import logging
import subprocess
from contextlib import contextmanager
from os import path

logger = logging.getLogger('connections')


@contextmanager
def set_argument(parsed_args, name, value):
    old_value = getattr(parsed_args, name)
    setattr(parsed_args, name, value)
    try:
        yield parsed_args
    finally:
        setattr(parsed_args, name, old_value)


def check_output_as_list(command):
    logger.debug(command)
    output = subprocess.check_output([command], shell=True)
    logger.debug(output)
    lines = output.decode('utf-8').split('\n')
    lines = [line for line in lines if len(line) > 0]
    logger.debug(lines)
    return lines


class Connection(object):
    @classmethod
    def hosts(cls, parsed_args):
        """ Returns a list of hosts to connect to.

        Can return a host as '\n' to force tmux to make a new window.
        """
        raise NotImplementedError()

    @classmethod
    def copy(cls, host, parsed_args):
        raise NotImplementedError()

    @classmethod
    def command(cls, host, parsed_args):
        raise NotImplementedError()

    @classmethod
    def connect(cls, host, parsed_args):
        raise NotImplementedError()


class SSHConnection(Connection):
    @classmethod
    def hosts(cls, parsed_args):
        if len(parsed_args.hosts) == 0:
            raise ValueError("At least one host must be specified!\n")
        return parsed_args.hosts

    @classmethod
    def copy(cls, host, parsed_args, and_execute=True):
        script_path, script_name = path.split(parsed_args.script)
        copy = 'scp {} {} {}:/tmp'.format(parsed_args.ssh_options, parsed_args.script, host)
        connect = cls.connect(host, parsed_args)
        chmod = connect + ' chmod u+x /tmp/{}'.format(script_name)
        if and_execute:
            execute = connect + ' /tmp/{}'.format(script_name)
            return '{} && {} && {} && {}'.format(copy, chmod, execute, connect)
        return '{} && {}'.format(copy, chmod)

    @classmethod
    def command(cls, host, parsed_args):
        return 'ssh {} {} {}'.format(parsed_args.ssh_options, host, parsed_args.command)

    @classmethod
    def connect(cls, host, parsed_args):
        return 'ssh {} {}'.format(parsed_args.ssh_options, host)


class DockerConnection(Connection):
    @classmethod
    def hosts(cls, parsed_args, prepend_command=''):
        host_names = parsed_args.hosts
        matched_host_names = {}
        for name in host_names:
            matched_host_names[name] = False
        hosts = []
        names_and_ids = check_output_as_list(
            prepend_command + "docker ps --format '{{.Names}},{{.ID}}'")
        if len(host_names) == 0:
            hosts.extend(n_and_i.split(',')[1] for n_and_i in names_and_ids)
        else:
            for n_and_i in names_and_ids:
                n, i = n_and_i.split(',')
                if n in host_names:
                    hosts.append(i)
                    matched_host_names[n] = True
                elif i in host_names:
                    hosts.append(i)
                    matched_host_names[n] = True
                elif parsed_args.approximate and any(name in n for name in host_names):
                    hosts.append(i)
                elif parsed_args.approximate and any(name in i for name in host_names):
                    hosts.append(i)

        logger.debug("hosts = {0}".format(hosts))

        if len(hosts) == 0:
            raise ValueError("No docker containers detected to connect to!")
        if not parsed_args.approximate:
            for name in host_names:
                if not matched_host_names[name]:
                    raise ValueError("No container named '{}' found!".format(name))

        return hosts

    @classmethod
    def _execute(cls, host, parsed_args, command, prepend_command=''):
        with set_argument(parsed_args, 'docker_command', 'exec -it {}'.format('{} ' + command)) as parsed_args:
            return cls.connect(host, parsed_args, prepend_command)

    @classmethod
    def copy(cls, host, parsed_args, prepend_command=''):
        script_path, script_name = path.split(parsed_args.script)
        copy = '{}docker cp {} {}:/tmp'.format(prepend_command, parsed_args.script, host)
        chmod = cls._execute(host, parsed_args, 'chmod u+x /tmp/' + script_name, prepend_command)
        execute = cls._execute(host, parsed_args, '/tmp/' + script_name, prepend_command)
        connect = cls.connect(host, parsed_args, prepend_command)
        return '{} && {} && {} && {}'.format(copy, chmod, execute, connect)

    @classmethod
    def command(cls, host, parsed_args, prepend_command=''):
        command = cls._execute(host, parsed_args, parsed_args.command, prepend_command)
        connect = cls.connect(host, parsed_args, prepend_command)
        return '{} && {}'.format(command, connect)

    @classmethod
    def connect(cls, host, parsed_args, prepend_command=''):
        if parsed_args.docker_command:
            if '{}' in parsed_args.docker_command:
                # Commands like:
                #  -dc 'exec -it {} bash'
                return '{}docker {}'.format(prepend_command, parsed_args.docker_command.format(host))
            else:
                # Commands without {} get host appended to it, like:
                #  -dc logs
                #  -dc 'logs -f'
                return '{}docker {} {}'.format(prepend_command, parsed_args.docker_command, host)
        else:
            return '{}docker exec -it {} bash'.format(prepend_command, host)


class DockerComposeConnection(DockerConnection):
    @classmethod
    def hosts(cls, parsed_args):
        containers = check_output_as_list('docker-compose ps --filter="status=running" --services')
        logger.debug('containers = "{}"'.format(containers))

        filtered_hosts = []
        for container_name in containers:
            if container_name in parsed_args.hosts:
                filtered_hosts.append(container_name)
            elif parsed_args.approximate and any(h in container_name for h in parsed_args.hosts):
                filtered_hosts.append(container_name)

        logger.debug('filtered_hosts = "{}"'.format(filtered_hosts))
        if len(filtered_hosts) == 0:
            raise ValueError("No service found in {}".format(containers))

        hosts = []
        for name in containers:
            if len(filtered_hosts) > 0 and name not in filtered_hosts:
                continue
            container = check_output_as_list('docker-compose ps -q {}'.format(name))
            if len(container) == 1:
                hosts.append(container[0])

        if len(hosts) == 0:
            raise ValueError("No running docker containers detected to connect to!")

        return hosts


class SSHDockerConnection(DockerConnection):
    @classmethod
    def hosts(cls, parsed_args):
        if len(parsed_args.hosts) == 0:
            raise ValueError("At least one host must be specified!\n")

        ssh_hosts = parsed_args.hosts

        with set_argument(parsed_args, 'hosts', parsed_args.docker_containers.split(',')) as parsed_args:
            hosts = []
            for ssh_host in ssh_hosts:
                found_containers = super().hosts(parsed_args, 'ssh {} '.format(ssh_host))
                more_hosts = ['{},{}'.format(ssh_host, container) for container in found_containers]
                if len(hosts) > 0 and len(more_hosts) > 0:
                    hosts.append('\n')
                hosts.extend(more_hosts)

            return hosts

    @classmethod
    def copy(cls, host, parsed_args):
        ssh_host, container = host.split(',')
        ssh_copy = SSHConnection.copy(ssh_host, parsed_args, and_execute=False)
        with set_argument(parsed_args, 'ssh_options', '-t ' + parsed_args.ssh_options) as parsed_args:
            ssh_command = SSHConnection.connect(ssh_host, parsed_args)
        with set_argument(parsed_args, 'script', '/tmp/' + parsed_args.script) as parsed_args:
            docker_copy = DockerConnection.copy(container, parsed_args, prepend_command=ssh_command + ' ')
        return '{} && {}'.format(ssh_copy, docker_copy)

    @classmethod
    def command(cls, host, parsed_args, prepend_command=''):
        ssh_host, container = host.split(',')
        with set_argument(parsed_args, 'ssh_options', '-t ' + parsed_args.ssh_options) as parsed_args:
            ssh_command = SSHConnection.connect(ssh_host, parsed_args)
            return DockerConnection.command(container, parsed_args, prepend_command=ssh_command + ' ')

    @classmethod
    def connect(cls, host, parsed_args):
        ssh_host, container = host.split(',')
        with set_argument(parsed_args, 'ssh_options', '-t ' + parsed_args.ssh_options) as parsed_args:
            ssh_command = SSHConnection.connect(ssh_host, parsed_args)
            return DockerConnection.connect(container, parsed_args, prepend_command=ssh_command + ' ')



import argparse
import logging
import posix
import sys

from . import tmux

logger = logging.getLogger('intmux')


def add_docker_options(subparser, include_hosts=True):
    subparser.add_argument(
        '--docker-command', '-dc', default='exec -it {} bash',
        help=(
            "Docker command to execute (default: 'exec -it {} bash'). If '{}' "
            "is included in the command, the docker host is substituted there, "
            "the host is appended. NOTE: may invalidate any --script/--command "
            "parameters if a shell is not provided."))
    subparser.add_argument(
        '--approximate', '-a', action='store_true',
        help='Include any docker container names that only partially match hosts.')
    if include_hosts:
        subparser.add_argument(
            'hosts', nargs='*',
            help=('List of docker containers to connect to (default: connect to all containers)'))


def add_ssh_options(subparser):
    subparser.add_argument(
        '--ssh-options', '-so', default="", help="Options to pass to SSH connection.")
    subparser.add_argument('hosts', nargs='*', help="SSH hosts to connect to.")


def main():
    parser = argparse.ArgumentParser(
        description="Connect to several hosts in a tmux session."
    )

    parser.add_argument(
        '--log', '-l', default="WARN",
        help="Log level (default: WARN)")
    parser.add_argument(
        '--command', '-c', default="",
        help="Command to execute when connecting to a remote host")
    parser.add_argument(
        '--input', '-i', type=argparse.FileType('r'), default=None,
        help="Read list of hosts from input file when provided.")
    parser.add_argument(
        '--script', '-s', default="",
        help="Execute commands in local file remotely (executes over --command option)")

    parser.add_argument(
        '--tmux-panes', '-p', default=6, type=int, metavar="PANES",
        help="Max tmux panes per window (default: 6)")
    parser.add_argument(
        '--tmux-no-sync', '-S', action='store_true',
        help="Do not run tmux's set-option synchronize-panes")
    parser.add_argument(
        '--tmux-session', '-t', default='intmux', metavar="SESSION",
        help="tmux session name (default: intmux)")

    subparsers = parser.add_subparsers(help='sub-command help', dest='subcommand')

    ssh_parser = subparsers.add_parser(
        'ssh', help='Connect to hosts via SSH',
        description='Connect to the provided hosts (or read from STDIN).')
    add_ssh_options(ssh_parser)

    docker_parser = subparsers.add_parser(
        'docker', help="Connect to docker containers via 'docker exec'",
        description='Connect to the provided running containers (or read from STDIN).')
    add_docker_options(docker_parser)

    ssh_docker_parser = subparsers.add_parser(
        'ssh-docker', help="Connect to docker containers on remote SSH hosts",
        description='Connect to docker containers on provided SSH hosts (or read from STDIN).')
    add_ssh_options(ssh_docker_parser)
    ssh_docker_parser.add_argument(
        '--docker-containers', '-dC',
        help=(
            'Comma separated list of docker containers to connect to '
            '(default: connect to all containers)'))
    add_docker_options(ssh_docker_parser, include_hosts=False)

    composer_parser = subparsers.add_parser(
        'compose', help="Connect to docker containers associated with current docker-compose via 'docker exec'",
        description=(
            'Connect to containers associated with the docker-compose '
            'in the current directory.'))
    add_docker_options(composer_parser)

    args = parser.parse_args()
    logging.basicConfig(level=getattr(logging, args.log))

    if not args.subcommand:
        print('You must supply a subcommand.')
        sys.exit(posix.EX_USAGE)

    session = tmux.TmuxSession(args)
    session.connect()



import logging
import os.path
import posix
import subprocess
import sys

from . import connections

logger = logging.getLogger('tmux')


def tmux(command):
    """ Run a tmux command """
    command = "tmux {}".format(command)
    logger.debug('tmux "{}"'.format(command))
    subprocess.check_call([command], shell=True)


def new_session(session):
    """ Create a new tmux session"""
    sessions = connections.check_output_as_list('tmux list-sessions -F "#S"')
    if session in sessions:
        print("Session '{}' already exists!".format(session))
        sys.exit(posix.EX_USAGE)

    command = "tmux new-session -d -s '{}'".format(session)
    logger.debug('new_session "{}"'.format(command))
    subprocess.check_call([command], shell=True)


class TmuxSession(object):

    def __init__(self, args):
        self.args = args
        self.session = args.tmux_session
        self.panes = args.tmux_panes
        self.sync = not args.tmux_no_sync
        self.command = args.command
        self.script = args.script

        if args.script and not os.path.exists(args.script):
            print("{} does not exist!".format(args.script))
            sys.exit(posix.EX_USAGE)

        if args.subcommand == 'ssh':
            self.connection_type = connections.SSHConnection
        elif args.subcommand == 'docker':
            self.connection_type = connections.DockerConnection
        elif args.subcommand == 'compose':
            self.connection_type = connections.DockerComposeConnection
        elif args.subcommand == 'ssh-docker':
            self.connection_type = connections.SSHDockerConnection
        else:
            print('Unknown subcommand type!')
            sys.exit(posix.EX_USAGE)

        self.hosts = []
        # Read hosts from stdin
        if not sys.stdin.isatty():
            for line in sys.stdin.readlines():
                self.hosts.append(line[:-1])
            logger.debug('STDIN hosts = {}'.format(self.hosts))
        elif args.input:
            for line in args.input.readlines():
                self.hosts.append(line[:-1])
            logger.debug('--input hosts = {}'.format(self.hosts))
        else:
            try:
                self.hosts = self.connection_type.hosts(args)
            except ValueError as e:
                print(e)
                sys.exit(posix.EX_USAGE)
            logger.debug('connection hosts = {}'.format(self.hosts))

        if len(self.hosts) == 0:
            print("At least one host must be specified!\n")
            sys.exit(posix.EX_USAGE)

    def connect(self):
        new_session(self.session)

        # turn on window activity notification:
        tmux("set-window-option -t {} -g monitor-activity on".format(self.session))
        tmux("set-option -t {} -g visual-activity on".format(self.session))

        cnt = 0
        wcnt = 0
        first = 1
        made_new_window = True
        for host in self.hosts:
            logger.debug('Host = {}'.format(host))
            # if hostname is '\n' then make a new window
            if host != '\n' and (cnt < self.panes or self.panes == 0):
                if first == 0:
                    tmux("split-window -t {}".format("{}:{}".format(self.session, wcnt)))
                first = 0
                cnt = cnt + 1
            else:
                if made_new_window and self.sync:
                    tmux("set-option -t {}:{} synchronize-panes".format(self.session, wcnt))
                    made_new_window = False
                made_new_window = True
                wcnt = wcnt + 1
                cnt = 1
                tmux("new-window -t {}".format(self.session))
                tmux("rename-window -t {}:{} {}".format(self.session, wcnt, host))
                tmux("set-window-option -t {}:{} allow-rename off".format(self.session, wcnt))

            if self.script:
                tmux("send-keys -t {}:{} \"{}\" C-m".format(
                    self.session, wcnt, self.connection_type.copy(host, self.args)))
            elif self.command:
                tmux("send-keys -t {}:{} \"{}\" C-m".format(
                    self.session, wcnt, self.connection_type.command(host, self.args)))
            else:
                tmux("send-keys -t {}:{} \"{}\" C-m".format(
                    self.session, wcnt, self.connection_type.connect(host, self.args)))

            tmux("select-layout -t {}:{} tiled".format(self.session, wcnt))

        if made_new_window and self.sync:
            logger.debug('synchronizing last window')
            tmux("set-option -t {}:{} synchronize-panes".format(self.session, wcnt))

        # remove session 0 - which is not connected to anything
        # TODO provide a hotkey to run in all sessions

        # Detect if we are already in a session. If we are, just switch to the other
        # session:
        if 'TMUX' in os.environ:
            # When quitting out of this session, just switch to some other client
            # (since there appears to be one already)
            tmux("set-option -g detach-on-destroy off")
            tmux("switch-client -t {}:{}".format(self.session, wcnt))
        else:
            tmux("attach-session -t {}:{}".format(self.session, wcnt))
